--<HTML>
--<HEAD>
--<Title>Library: EuCANOOP.e</TITLE>
--<LINK HREF=eustyle.css REL="STYLESHEET" TYPE="text/css">
--</HEAD>
--<BODY>
--<div id="site_header">
-- <div id="site_heading"><a href="http://sourceforge.net/projects/eucanoop/">
-- EuCANOOP v3.0.0</a></div> 
-- <div>OOP library for use with the <a href="http://www.rapideuphoria.com/">
-- Euphoria Programming Language</a></div> 
-- </div> 

--<h1>EuCANOOP</h1>

--<h2>Eu[phoria] can [do] OOP</h2>

-- OOP programming in Euphoria is not built into the language.
-- It is, however, quite simple to achieve and easy to apply.
-- The library in EuCANOOP provides a means of carrying out OOP
-- with just nine routines.

--<p>
-- "Standard" Euphoria manipulates data structures such as
-- <code>object</code>s, <code>sequence</code>s, <code>atom</code>s,
-- <code>integer</code>s and any other type the user chooses to adopt
-- or define. It does the manipulation by means of routines, some
-- written in the program, others being in libraries.
-- Authors might be RDS, the user or, perhaps, other users.
--</p>
--<p>
-- EuCANOOP offers an exactly analogous form of processing, but
-- with just one data structure - the Object - and, at this stage
-- anyway, just two pre-written routines.
-- The difference is that the variables and routines are stored
-- within the data structure and manipulation takes place solely
-- within the Object.</p>

--<p>
-- You start by defining Classes, as prototypes of Objects.
-- After each definition is complete the details of that Class
-- are stored for future use. Once you have declared a Class you
-- can create Instances of it (Objects), setting values specific
-- to the Instance. Manipulation of an Object is possible because
-- it carries within its values not only its "variables" but also
-- pointers to routines (both <code>procedures</code> and
-- <code>functions</code>) which operate on the Object's variables.</p>

--<p>This is all achieved using two standard Euphoria constructs:</p>
--<ol>
--<li>the <code>sequence</code>
--<li>the function <code>routine_id</code>
--</ol>

--<p>The rest is what I once heard called "syntactical sugar" -
-- coding to provide stylistic simplification.
-- This is what the EuCANOOP library provides.

--<p> Read on if you wish for more detail now,
-- or jump to the library's <a href=#Spec>specification</a>.

--<h3>The library described</h3>

--<p>
-- You define a Class using the procedure <code>Class</code>.
-- It takes one parameter:
--<ul>
--<li>a <code>string</code> containing the name you wish to associate
-- with that Class.
--</ul>

--<pre class=snip>Class("Dog")</pre>

-- If you wish the Class to inherit from another pre-defined
-- Class then use the procedure <code>Extends</code>.
-- Multiple inheritance is supported.

-- <code>Extends</code> takes one parameter:
--<ul>
--<li>an <code>array</code> of one or more <code>string</code>s
-- containing the name(s) of the Class(es).
--</ul>

--<pre class=snip>Extends({"Animal"})</pre>

--<p>
-- If you wish to add property values then you use the
-- procedure <code>Property</code>. It takes three parameters:</p>
--<ul>
--<li>a <code>string</code> containing the name you wish to associate
-- with the property.
--<li>an <code>integer</code> signifying the type of the variable
-- (ATOM for <code>atom</code>, SEQUENCE for <code>sequence</code>, etc).
--<li>an <code>object</code> containing the default value for the property.
-- This value should be of the correct type. If it isn't then a basic default
-- is set instead.
--</ul>

--<p>
-- My convention is to use upper-case text for property names.</p>

--<pre class=snip>Property("FEET",INTEGER,0)</pre>

--<p>
-- If you wish to add a method then use the procedure <code>Method</code>.
-- It takes two parameters:
--<ul>
--<li>a <code>string</code> containing the name you wish to associate
-- with the method.
--<li>an <code>integer</code> carrying the return value from
-- <code>routine_id</code>.
-- The parameter for that call should be the name of the routine you wish
-- to invoke when the method is called
-- (see later for how to code methods and how to invoke them).
-- The routine is either pre-defined, outside the Class definition, or
-- may be included within it.
--</ul>

-- My convention is to use title-case for method names.

--<pre class=snip>Method("Bark",routine_id("Dog_Bark"))</pre>

-- Repeat Property and/or Method calls as required to define all the values
-- you need.

--<p>
-- If you wish to assign an inherited Method to a different routine then
-- you use the procedure <code>Override</code>.
-- Just like <code>Method</code> it takes two parameters:
--<ul>
--<li>a <code>string</code> containing the name of the inherited method.
--<li>an <code>integer</code> carrying the return value from
-- <code>routine_id</code> for the "new" routine.
--</ul>

--<p>
-- A Class definition is only stored when the procedure
-- <code>EndClass</code> is called. It takes one parameter:
--<ul>
--<li>a <code>string</code> containing the name you have already used to
-- define the Class.
--</ul>

--If you fail to include such a call, or key in the wrong name then the
-- Class will not be entered into the database of Classes.

--<pre class=snip>EndClass("Dog")</pre>

--I find it useful to use an indenting convention for Class declaration.
-- For example:

--<pre class=list>Class("Dog")
--    Extends({"Animal"})
--    Property("BREED",SEQUENCE,"")
--    Method("Bark",routine_id("Dog_Bark"))
--EndClass("Dog")
--</pre>

-- EuCANOOP's Objects are operational instances of Class paradigms.
-- Objects are created (in the image of their Class definition) using
-- the function <code>New</code>. It takes two parameters:
--<ul>
--<li>a <code>string</code> containing the name of the Class from which
-- the Object derives.
--<li>an <code>array</code> of one or more <code>array</code>s.
-- Each inner array contains two elements:
--<ul>
--<li>a <code>string</code> containing the name of the property whose
-- value you wish to set.
--<li>an <code>object</code> containing the value you wish to assign
-- to that property.
--(If the value is not of the correct type then the default setting for the
-- class is substituted.)
--</ul>
--</ul>

--You only need include such an array if you want to change from the default[s].

--<pre class=snip>Fido = New("Dog",{{"BREED","Mongrel"}})</pre>

--All Objects are created with two standard Methods:
-- <code>Get</code> and <code>Set</code>.

--These Methods can be used to inspect or change the values of properties.
--<p>Methods are called as follows:
--<p>If the Method relates to a <code>procedure</code> then you use
--<code>MethodProc</code>. It takes three parameters:
--<ul>
--<li>[<code>Instance</code>] name of the Object.
--<li>a <code>string</code> containing the name of the Method.
--<li>a <code>sequence</code> containing the parameters required by the
--procedure associated with the method - excluding the first (see later).
--</ul>

--<pre class=snip>MethodProc(StdOut,"Puts",{"Hello, World!"})</pre>

--<p>If the Method relates to a <code>function</code> then you use
--<code>MethodFunc</code>. It takes three parameters:
--<ul>
--<li>[<code>Instance</code>] name of the Object.
--<li>a <code>string</code> containing the name of the Method.
--<li>a <code>sequence</code> containing the parameters required by the
--procedure associated with the method - excluding the first (see later).
--</ul>

--<pre class=snip>if MethodFunc(StdIn,"GetC",{}) then end if</pre>

--Routines (qua Methods) must be defined as follows:
--<p>The routine must have at least one parameter. This is an
--<code>Instance</code>, which is used to refer to the Object itself.
--The remaining parameters, if any, are the ones to be called by the Method.
--<p>To avoid ambiguity my convention is to name routines in a consistent
--fashion. For example, if I intend to define a method called "What", 
--for Class "Fred", with two method parameters (excluding "self") then I
--would name the routine <code>Fred_What_2</code>.

--<p>Example:
--<pre class=snip>procedure Output_Puts_1(Instance self,string text)
--    puts(self,text)
--end procedure
--</pre>

--<h3>Constructors</h3>


-- In some circumstances calling <code>New</code> is insufficient.
-- Moreover, some users with other OOP language experience actually want to
-- define Constructors. You can easily write conventional constructors by
-- combining a call to <code>New</code> with a relevant creator
-- routine. So, for example:


--<pre class=snip>constant myFile = New("File",{{"NAME","Hello.ex"},{"MODE","r"}})
--MethodProc(myFile,"Open",{})
--</pre>


--could be replaced by

--<pre class=snip>constant myFile = File("Hello.ex","r")</pre>


--if the constructor <code>File</code> is defined as follows:


--<pre class=snip>global function File(string name,string mode)
--    Instance obj
--    integer handle
--    obj = New("File",{{"NAME",name},{"MODE",mode}})
--    handle = open(name,mode)
--    MethodProc(obj,"Set",{"HANDLE",handle})
--    return obj
--end function
--</pre>

-- My convention is to use, as the creator function, a function with precisely
-- the same name as the Class.

--<h3>Destructors</h3>

-- Because Euphoria has built-in garbage collection, there is no real need
-- to define destructor routines. This even applies to operations such as
-- closing an already opened file - Euphoria automatically closes all files
-- when an application terminates.

--<h3>Types</h3>

-- Objects are all declared to be of type <code>Instance</code>, which enables
-- the programmer to distinguish all his/her Object instances.

--<p>
-- This version extends to range of possible types - both for general
-- programming use and in defining EuCANOOP properties.
-- The additional types offered are:
--<ul>
--<li> <code>array</code> - a sequence where the types of all elements are
--     mutually compatible
--<li> <code>bool</code> - TRUE or FALSE
--<li> <code>byte</code> - a value in the range 0-255
--<li> <code>character</code> - a value in the range 0-127
--<li> <code>string</code> - an array of bytes
--<li> <code>array_of_strings</code>
--     - an array in which all elements are strings
--</ul>

--<p>
-- Just like Euphoria's variables, properties of Objects are typed. EuCANOOP,
-- however, uses its own type-checking routines, so property definitions use
-- EuCANOOP's own types. In addition to the base types defined in Version 2
-- (OBJECT, ATOM, INTEGER and SEQUENCE) this version offers the addditional
-- EuCANOOP types:
-- ARRAY, BOOL, BYTE, CHARACTER, STRING and STRING_ARRAY. 

--<p>
-- Type-checking in EuCANOOP takes place when a Class Property's default is
-- set, when an <code>Instance</code> is created with a non-default parameter
-- and when the built-in "Set" Method is invoked. Unlike Euphoria, a type error
-- does not result in program termination. Although an error message is issued
-- if an assignment fails, the property's value is set, respectively, to a
-- "null" value for that type, the Class default or is left unchanged.

--<p>
-- Just as in standard Euphoria it is easy to define your own EuCANOOP types
-- to add to the list above. To find out how see <a href="adding_types.html">
-- here</a>.

--<h2><a name=Spec></a>Library specification</h2>

--<h3>Version: 3.0.0</h3>
--<dir>
--     Date: November 2009
--<br> Author: C A Newbould (canewbould@users.sourceforge.net)
--<br> Changes:
--<ul>
--<li> Types added (for use in Euphoria coding or for specifying EuCANOOP
--     Properties)
--<li> Documentation style changed
--</ul>
--</dir>

--<h2>Interface</h2>

  --<h3>Globally-defined includes:</h3>
  --<ul>
  --<li>types.e - for Eu4's extended types
  --</ul>

  --Locally-defined includes:
  -- include std/io.e (Eu4 only)
  --<li>io.e - for I/O functionality

--<h3>Constants:</h3>
--<ul>
global constant TRUE = 1,
--<li>TRUE - set to 1
                FALSE = 0,
--<li>FALSE - set to 0
                EMPTY_STRING = ""       
--<li>EMPTY_STRING = ""
--</ul>

--<h3>Routines:</h3>
  -- local routines
function get_type(object obj)
    sequence mask
    mask = repeat(FALSE,3)
    if atom(obj) then
        mask[1] = TRUE
    end if
    if integer(obj) then
        mask[2] = TRUE
    end if
    if sequence(obj) then
        mask[3] = TRUE
    end if
    return mask
end function

function iif(integer expression,object true,object false)
    if expression then
        return true
    else
        return false
    end if
end function

-- ** Routines for type-checking of properties **
--<ul>  
global function is_array(object a)
	object first
	integer len
	integer	ret
	if not sequence(a) then
		return FALSE
	else
		ret = TRUE
		len = length(a)
		if len > 0 then
			first = get_type(a[1])
			for i = 2 to len do
				if not equal(get_type(a[i]),first) then
					ret = FALSE
				end if
			end for
		end if
	end if
	return ret
end function
--<li>[function] is_array(<code>object</code> a)
--    - checks that <code>a</code> is an <code>array</code>
global function is_atom(object a)
    if not atom(a) then
        return FALSE
    else
        return atom(a)
    end if
end function
--<li>[function] is_atom(<code>object</code> a)
--    - checks that <code>a</code> is an <code>atom</code>
global function is_bool(object bool)
	-- A bool is a value that is either zero or not zero.
    if not integer(bool) then
        return FALSE
    else
        if (bool = TRUE or bool = FALSE) then
            return TRUE
        else
            return FALSE
        end if
    end if
end function
--<li>[function] is_bool(<code>object</code> bool)
--    - checks that <code>bool</code> is a <code>bool</code>
global function is_byte(object byt)
    if not integer(byt) then
        return FALSE
    else
        if byt >= 0 and byt < 256 then
            return TRUE
        else
            return FALSE
        end if
    end if
end function
--<li>[function] is_byte(<code>object</code> byt)
--    - checks that <code>byt</code> is a <code>byte</code> (0 to 255)
global function is_character(object char)
    if not integer(char) then
        return FALSE
    else
        if char >= 0 and char < 128 then
            return TRUE
        else
            return FALSE
        end if
    end if
end function
--<li>[function] is_character(<code>object</code> char)
--    - checks that <code>char</code> is a <code>character</code> (0 to 127)
global function is_instance(object i)
    integer inst
    inst = FALSE
    if not integer(i) then
        inst = FALSE
    else
        if i > -2 then
            inst = TRUE
        end if
    end if
    return inst
end function
--<li>[function] is_instance(<code>object</code> a)
--    - checks that <code>a</code> is an <code>Instance</code>
global function is_integer(object i)
    if not integer(i) then
        return FALSE
    else
        return integer(i)
    end if
end function
--<li>[function] is_integer(<code>object</code> a)
--    - checks that <code>a</code> is an <code>integer</code>
global function is_object(object o)
    return TRUE
end function
--<li>[function] is_object(<code>object</code> a)
--    - checks that <code>a</code> is an <code>object</code>
global function is_sequence(object s)
    if not sequence(s) then
        return FALSE
    else
        return sequence(s)
    end if
end function
--<li>[function] is_sequence(<code>object</code> a)
--    - checks that <code>a</code> is a <code>sequence</code>
global function is_string(object str)
    integer result
    if not is_array(str) then
        result = FALSE
    else
        result = TRUE
        for i = 1 to length(str) do
            if not is_byte(str[i]) then
                result = FALSE
            end if   
        end for
    end if
    return result
end function
--<li>[function] is_string(<code>object</code> str)
--    - checks that <code>str</code> is a <code>string</code>
global function is_array_of_strings(object arr)
    integer result
    if not is_array(arr) then
        result = FALSE
    else
        result = TRUE
        for i = 1 to length(arr) do
            if not is_string(arr[i]) then
                result = FALSE
            end if
        end for
    end if
    return result
end function
--<li>[function] is_array_of_strings(<code>object</code> str)
--    - checks that <code>str</code> is an <code>array</code> of
-- <code>string</code>s
function is_array_of_integers(object arr)
    integer result
    if not is_array(arr) then
        result = FALSE
    else
        result = TRUE
        for i = 1 to length(arr) do
            if not integer(arr[i]) then
                result = FALSE
            end if
        end for
    end if
    return result
end function

--</ul>

--<h3>Types:</h3>
--<ul>
global type array(sequence arr)
    return is_array(arr)
end type
--<li><code>array</code>
-- - a one-dimensional array of compatible type (all <code>atom</code>,
-- all <code>integer</code> or all <code>sequence</code>)
--<br> (There is a slight problem here - Euphoria's <code>integer</code>
-- function doesn't distinguish between x and x.0!)
global type bool(object b)
	return is_bool(b)
end type
--<li><code>bool</code> - TRUE or FALSE
global type byte(object byt)
    return is_byte(byt)
end type
--<li><code>byte</code> - 0 to 255
global type character(object char)
    return is_character(char)
end type
--<li><code>character</code> - 0 to 127
global type Instance(object inst)
    return is_instance(inst)
end type
--<li><code>Instance</code> - for instantiation of Objects
global type string(sequence str)
    return is_string(str)
end type
--<li><code>string</code> - a one-dimensional array of characters
-- including the empty string {}
global type string_array(sequence arr)
   return is_array_of_strings(arr)
end type
--<li><code>string_array</code> - a one-dimensional array of strings
--</ul>

  -- Local type
global type integer_array(sequence arr)
   return is_array_of_integers(arr)
end type  

--<h3> EuCANOOP types:</h3>
--<ul>
global constant ARRAY = routine_id("is_array"),
--<li>ARRAY
                ATOM = routine_id("is_atom"),
--<li>ATOM
                BOOL = routine_id("is_bool"),
--<li>BOOL
                BYTE = routine_id("is_byte"),
--<li>BYTE
                CHARACTER = routine_id("is_character"),
--<li>CHARACTER
                INSTANCE = routine_id("is_instance"),
--<li>INSTANCE
                INTEGER = routine_id("is_integer"),
--<li>INTEGER
                OBJECT = routine_id("is_object"),
--<li>OBJECT
                METHOD = -999,
--<li>METHOD - to signal that the "Property" is, in fact, a Method
                SEQUENCE = routine_id("is_sequence"),
--<li>SEQUENCE
                STRING = routine_id("is_string"),
--<li>STRING
                STRING_ARRAY = routine_id("is_array_of_strings")
--<li>STRING_ARRAY
--</ul>

  --<h3>Constants:</h3>
  -- pointers within an Object
constant
    TYP_ERR = "**ERROR** Type conflict in Property '%s' when ",

  --enum PROPERTY, VALUE, TYPE (Eu4)
    PROPERTY = 1,
	VALUE = 2,
	TYPE = 3,
  --defined in Eu4 through io.e
	SCREEN = 1,
	STDERR = 2
	
  --<h3>Variables:</h3>
sequence Classes,
         Objects

string_array ClassList

string className
string_array properties
sequence values
integer_array types
integer objectCounter

  -- Initialisation
objectCounter = 0
Classes = {}
ClassList = {}
Objects = {}

--<h3>Routines:</h3>

  -- Local routines

procedure AddClass()
    if length(ClassList) = 0 then
        ClassList = {className}
        Classes = {{properties,values,types}}
    else
        ClassList = append(ClassList,className)
        Classes = append(Classes,{properties,values,types})
    end if
end procedure

function CheckType(integer typ,object val)
    return call_func(typ,{val})
end function

function FindClass(sequence class)
    integer found
    found = find(class,ClassList)
    if found = 0 then
        printf(STDERR,"\n**ERROR** Class: '%s' not found!\n",{class})
    end if
    return found
end function

function FindClassProperty(string class,string property)
    integer cl,
            found
    string_array props
    cl = FindClass(class)
    props = Classes[cl][PROPERTY]
    found = find(property,props)
    if found = 0 then
        printf(STDERR,
               "\n**ERROR** Class %s: '%s' not found!\n",
               {{class},{property}})
    end if
    return found
end function

function FindObjectProperty(Instance self,string property)
    integer found
    string_array props
    props = Objects[self][PROPERTY]
    found = find(property,props)
    if found = 0 then
        printf(STDERR,
               "\n**ERROR** Object #%d: '%s' not found!\n",
               {{self},{property}})
    end if
    return found
end function

function GetProperty(Instance self,string property)
    return Objects[self][VALUE][FindObjectProperty(self,property)]
end function

procedure SetProperty(Instance self,string property,object val)
    integer prop,
            typ
    prop = FindObjectProperty(self,property)
    typ = Objects[self][TYPE][prop]
    if typ = METHOD then
        printf(STDERR,"\n**ERROR** Attempt made to change Method \"%s\". ",
               {Objects[self][PROPERTY][prop]})
        puts(STDERR,"\n(Use \"Override\" function to do this).\n")
    else
        -- check type on each assignment
        if CheckType(typ,val) then
            Objects[self][VALUE][prop] = val
        else
            printf(STDERR,TYP_ERR & "using \"Set\" Method!\n",{property})    
        end if
    end if
end procedure

procedure Base()
    properties = {"Get","Set"}
    values = {routine_id("GetProperty"),routine_id("SetProperty")}
    types = {METHOD,METHOD}
end procedure

  -- Global routines
--** Routines for Class creation **
--<ul>
global procedure Class(string name)
    className = name
    properties = {}
    values = {}
    types = {}
end procedure
--<li>[procedure] <font color=purple>Class</font>(<code>string</code> name)
-- - opens the class definition
global procedure Extends(string_array parents)
    integer found,
            parent
    string_array props
    sequence vals
    integer_array typs
    if length(parents) = 0 then
        properties ={}
        values = {}
        types = {}
    elsif length(parents) >= 1 then
        parent = FindClass(parents[1])
        if parent > 0 then
            properties &= Classes[parent][PROPERTY]
            values &= Classes[parent][VALUE]
            types &= Classes[parent][TYPE] 
        end if
    end if
    if length(parents) > 1 then
        -- to avoid possibility of same property/method defined twice
        for i = 2 to length(parents) do
            parent = FindClass(parents[i])
            if parent > 0 then
                props = Classes[parent][PROPERTY]
                vals = Classes[parent][VALUE]
                typs = Classes[parent][TYPE]
                for j = 1 to length(props) do
                    found = find(props[j],properties)
                    if found = 0 then
                        properties = append(properties,props[j])
                        values = append(values,vals[j])
                        types = append(types,typs[j])
                    end if
                end for
            end if
        end for
    end if
end procedure
--<li>[procedure] <font color=purple>Extends</font>(<code>string_array</code>
-- parents)
-- - inherits from parent(s)
global procedure Property(string property,integer typ,object default)
    if CheckType(typ,default) then
        values = append(values,default)
    else
        if typ = SEQUENCE
        or typ = STRING
        then
            values = append(values,EMPTY_STRING)
        elsif typ = ARRAY then
            values = append(values,{0})
        elsif typ = STRING_ARRAY then
            values = append(values,{""})
        else
            values = append(values,0)
        end if
        printf(STDERR,TYP_ERR & "setting default!\n",{property})
    end if
    properties = append(properties,property)
    
    types = append(types,typ)
end procedure
--<li>[procedure] <font color=purple>Property</font>(<code>string</code>
-- property,
-- <code>integer</code> typ,<code>object</code> default) -
-- defines a new property, names it, declares its type, and specifies
-- a default value
global procedure Method(string method,integer rid)
    if rid > -1 then
        properties = append(properties,method)
        values = append(values,rid)
        types = append(types,METHOD)
    else
        printf(STDERR,"\n\n**ERROR** Method '%s' undefined!\n",{method})
    end if
end procedure
--<li>[procedure] <font color=purple>Method</font>(<code>string</code>
-- method,<code>integer</code> rid)
-- - defines a new method, names it and indicates the routine-id to which it
-- refers
global procedure Override(string method,integer rid)
    integer found
    found = find(method,properties)
    if found = 0 then
        printf(STDERR,"\n**ERROR** Method: '%s' not found!\n",{method})
    elsif rid < 0 then
        puts(STDERR,"\n**ERROR** Invalid routine-id!\n")
    else
        values[found] = rid
    end if
end procedure
--<li>[procedure] <font color=purple>Override</font>(<code>string</code> method,
-- <code>integer</code> rid) -
-- defines the routine-id to which an existing method should now refer
global procedure EndClass(string name)
    if equal(name,className) then
        AddClass()
    else
        printf(STDERR,"\n**ERROR** Class name conflict: '%s' is not '%s'!\n",
                 {name,className})
    end if
end procedure
--<li>[procedure] <font color=purple>EndClass</font>(<code>string</code> name)
-- - completes the class definition
--</ul>

procedure report_error(string text)
    puts(STDERR,text & " not defined!")
end procedure
  --<li>procedure report_error(string text)

--** Routines for Object manipuation **
--<ul>
global function MethodFunc(Instance self,string method,sequence args)
    integer rid
    rid = GetProperty(self,method)
    if rid >= 0 then
        if length(args) = 0 then
            return call_func(rid,{self})
        else
            args = prepend(args,self)
            return call_func(rid,args)
        end if
    else
        report_error("Method")
    end if
end function

--<li>[function] <font color=purple>MethodFunc</font>(<code>Instance</code>
-- self,<code>string</code> method,<code>sequence</code> args)
global procedure MethodProc(Instance self,string method,sequence args)
    integer rid
    rid = GetProperty(self,method)
    if rid >= 0 then
        if length(args) = 0 then
            call_proc(rid,{self})
        else
            args = prepend(args,self)
            call_proc(rid,args)
        end if
    else
        report_error("Method")
    end if
end procedure
--<li>[procedure] <font color=purple>MethodProc</font>(<code>Instance</code>
-- self,<code>string</code> method,<code>sequence</code> args)
global function New(string class,sequence args)
    integer cl,
            found
    Base()
    cl = FindClass(class)
    properties &= Classes[cl][PROPERTY]
    values &= Classes[cl][VALUE]
    types &= Classes[cl][TYPE]
    -- set parameters
    if length(args) > 0 then
        for i = 1 to length(args) do
            found = find(args[i][PROPERTY],properties)
            if found > 0 then
                if CheckType(types[found],args[i][VALUE]) then
                    values[found] = args[i][VALUE]
                else
                    printf(STDERR,TYP_ERR & "instantiating Object!\n",
                           {args[i][PROPERTY]})    
                end if      
            end if
        end for
    end if
    -- Add to Objects
    if objectCounter = 0 then
        Objects = {{properties,values,types}}
    else
        Objects = append(Objects,{properties,values,types})
    end if
    objectCounter += 1
    return objectCounter
end function
--<li>[function] <font color=purple>New</font>(<code>string</code>
-- class,<code>sequence</code> args)
-- - creates an Instance of type class, with the parameter to
-- initialise properties.
-- The args should be sequences containing pairs:
--<ul>
--<li>name of property
--<li>value to be assigned to that property
--</ul>
-- The return value is a pointer to the Object
--</ul>

--** Built-in Methods **
--<ul>
--<li>[function] <font color=purple>"Get"</font>
-- - returns the value of the specified Property
--<br>Syntax: <font color=purple>MethodFunc</font>(instance_name,
--<font color=purple>"Get"</font>,{property_name})
--<li>[procedure] <font color=purple>"Set"</font>
-- - changes the value of the specified Property
--<br>Syntax: <font color=purple>MethodProc</font>(instance_name,
--<font color=purple>"Set"</font>,{property_name,value})
--<br>If the value is not of the correct type then an error message is issued
--and the value remains unchanged.
--</ul>

--** Inspection Routines **
--<ul>
global procedure ListClasses()
    puts(SCREEN,"\n\n** List of Classes **")
    printf(SCREEN,"\n   At this point in time %d classes have been defined.",
           length(Classes))
    puts(SCREEN,"\n   These are: ")
    for i = 1 to length(Classes)-1 do
        puts(SCREEN,"\"" & ClassList[i] &"\"; ")
    end for
    puts(SCREEN,"\"" & ClassList[$] & "\".\n** End **\n")
end procedure

procedure report(string prompt,string_array props,sequence vals,integer_array typs)
    for i = 1 to length(props) do
        if typs[i] = METHOD then
            puts(1,"\nMethod")
        else
            puts(1,"\nProperty")
        end if
        printf(1,prompt,{props[i]})
        if typs[i] = STRING then
            printf(1," value \"%s\"",{vals[i]})
        elsif typs[i] = STRING_ARRAY then
            puts(1," value {")
            if length(vals[i]) > 1 then
                for j = 1 to length(vals[i]) - 1 do
                    printf(1,"\"%s\",",{vals[i][j]})
                end for
                printf(1,"\"%s\"}",{vals[i][$]})
            else
                printf(1,"\"%s\"}",{vals[i][$]})
            end if            
        elsif typs[i] = SEQUENCE
           or typs[i] = ARRAY
           or typs[i] = OBJECT and sequence(vals[i])
           then
            puts(1," value ")
            if length(vals[i]) = 0 then
                puts(1,"{}")
            else
                puts(1,"{")
                for j = 1 to length(vals[i]) - 1 do
                    print(1,vals[i][j])
                    puts(1,",")
                end for
                print(1,vals[i][$])
                puts(1,"}")
            end if
        elsif typs[i] = BOOL then
            printf(1," value %s",{iif(vals[i],"TRUE","FALSE")})        
        else
            printf(1," value %g",{vals[i]})
        end if
    end for
    puts(1,"\n** End **\n\n")
end procedure

--<li>[procedure] <font color=purple>ListClasses</font>()
-- - counts and lists the Classes defined
global procedure ListClass(string class)
    integer cl
    string_array props
    sequence vals
    string prompt
    integer_array typs
    cl = FindClass(class)
    props = Classes[cl][PROPERTY]
    vals = Classes[cl][VALUE]
    typs = Classes[cl][TYPE]
    printf(1,"\n\n** Details of Class: \"%s\" **",{class})
    prompt = " \"%s\":"
    report(prompt,props,vals,typs)
end procedure
--<li>[procedure] <font color=purple>ListClass</font>(<code>string</code> class)
-- - details the contents of the Class
global procedure ListObject(Instance obj)
    string_array props
    sequence vals
    string prompt
    integer_array typs
    props = Objects[obj][PROPERTY]
    vals = Objects[obj][VALUE]
    typs = Objects[obj][TYPE]
    printf(1,"\n\n** Details of Object: %d **",{obj})
    prompt = " \"%s\" has"
    report(prompt,props,vals,typs)
end procedure
--<li>[procedure] <font color=purple>ListObject</font>(<code>Instance</code>
-- obj)
-- - details the current contents
-- of the Object
--</ul>

--</BODY>
--</HTML>
