--<HTML>
--<HEAD>
--<Title>Library: IO</TITLE>
--<LINK HREF=eustyle.css REL="STYLESHEET" TYPE="text/css">
--</HEAD>
--<BODY>
--<div id="site_header">
--<font color = teal>
--<div>I/O library for use with the
--<a href="http://sourceforge.net/projects/eucanoop/">EuCANOOP library</a></div> 
--<div>EuCANOOP is an OOP library for use with the 
--<a href="http://www.rapideuphoria.com/">Euphoria Programming Language</a>
--</div>
--</font> 
--</div>

--<h1>IO.e</h1>
-- The library provides the Objects needed to use Input-Output in EuCANOOP.

--<p>All these Objects are <b>Devices</b> - they have a property which is the
-- "handle" by which they are identified.
-- The handle is either fixed (defined by Euphoria) or is generated using a
-- Constructor assignment. The Euphoria equivalent is to use a creation function
-- (for example, "open" for a file, or "create" in GUIs).
-- A return value of "-1" means that the creation function failed.

--<h2>Library specification</h2>
--<h3>Version: 2.1.0</h3>
--<dir>
--    Date: November 2008
--<br>Author: C A Newbould (canewbould@users.sourceforge.net)
--<br> Changes:
--<ul>
--<li> "Screen" Class name restored
--<li> Names of defined instances changed to reflect, but not conflict with,
-- those in Euphoria 4's standard "io" library.
--</ul>
--</dir>

--<h2>Interface</h2>

--<h3>Globally-defined includes:</h3>
--<ul>
include EuCANOOP.e
--<li>EuCANOOP.e - for OOP facilities
include get.e
--<li>get.e - for <code>get</code> function
--</ul>

  --Locally-defined includes:

--<h3>Constants:</h3>
--<ul>
global constant EOF = -1
--<li>EOF - end-of-file marker (set to -1)
      -- - defined in Eu4's io.e
--</ul>

--<h3>Classes:</h3>
--<ul>
Class("Device")
--<li><h3><a name=Device>"Device"</a></h3>
--<dir>
--<h4>Parents: --NONE--</h4>
--<h4>Properties:</h4>
--<ul>
--<li>HANDLE - <code>INTEGER</code> handle to the device
    Property("HANDLE",INTEGER,-1)
--</ul>
EndClass("Device")
--</dir>

Class("Output")
--<li><h3><a name=Output>"Output"</a></h3>
--<dir>
--<h4>Parents: <a href=#Device>"Device"</a></h4>
    Extends({"Device"})
--<h4>Methods:</h4>
--<ul>
	procedure Output_Puts(Instance self,sequence text)
        puts(MethodFunc(self,"Get",{"HANDLE"}),text)
	end procedure
    Method("Puts",routine_id("Output_Puts"))
--<li>[procedure] Puts(<code>sequence</code> text
-- - writes the text to the Output Device
	procedure Output_Print(Instance self,object obj)
        print(MethodFunc(self,"Get",{"HANDLE"}),obj)
	end procedure
    Method("Print",routine_id("Output_Print"))
--<li>[procedure] Print(<code>object</code> obj)
-- - prints the object to the Output Device
	procedure Output_Print_Format(Instance self,string format,object obj)
    	printf(MethodFunc(self,"Get",{"HANDLE"}),format,obj)
	end procedure
    Method("PrintFormat",routine_id("Output_Print_Format"))
--<li>[procedure] PrintFormat(<code>string</code> format,<code>object</code>
-- obj) - outputs the object to the Output Device in the given format
--</ul>
EndClass("Output")
--</dir>

Class("Screen")
--<li><h3><a name=Screen>"Screen"</a></h3>
--<dir>
--<h4>Parents: <a href=#Output>"Output"</a></h4>
    Extends({"Output"})
--<h4>Methods:</h4>
--<ul>
	procedure Screen_Clear(Instance self)
    	clear_screen()
	end procedure
	Method("Clear",routine_id("Screen_Clear"))
--<li>[procedure] Clear() - clears the screen
--</ul>
EndClass("Screen")
--</dir>

Class("Input")
--<li><h3><a name=Input>"Input"</a></h3>
--<dir>
--<h4>Parents: <a href=#Device>"Device"</a></h4>
    Extends({"Device"})
--<h4>Methods:</h4>
--<ul>
	function Input_GetA(Instance self)
    	sequence no
    	no = get(MethodFunc(self,"Get",{"HANDLE"}))
    	if no[1] = 0 then
        	return no[2]
    	elsif no[1] = GET_EOF then
        	return EOF
    	end if
	end function
    Method("GetA",routine_id("Input_GetA"))
--<li>[function] GetA() - reads an <code>atom</code> from the Input Device 
	function Input_GetB(Instance self)
	    object line
	    line = gets(MethodFunc(self,"Get",{"HANDLE"}))
	    if atom(line) then
	        return FALSE
	    else
	        if line[$] = 10 then
	            line = line[1..$-1]
	        end if
	        if equal(line,"TRUE") then
	            return TRUE
	        else
	            return FALSE
	        end if
	    end if
	end function
    Method("GetB",routine_id("Input_GetB"))
--<li>[function] GetB() - reads a <code>bool</code>ean value from the Input Device
	function Input_GetC(Instance self)
	    return getc(MethodFunc(self,"Get",{"HANDLE"}))
	end function
    Method("GetC",routine_id("Input_GetC"))
--<li>[function] GetC() - reads a <code>character</code>/<code>byte</code> from the Input Device
	function Input_GetI(Instance self)
	    sequence no
	    no = get(MethodFunc(self,"Get",{"HANDLE"}))
	    if no[1] = 0 then
	        return no[2]
	    elsif no[1] = GET_EOF then
	        return EOF
	    end if
	end function
    Method("GetI",routine_id("Input_GetI"))
--<li>[function] GetI() - reads an <code>integer</code> from the Input Device
	function Input_GetS(Instance self)
	    object line
	    line = gets(MethodFunc(self,"Get",{"HANDLE"}))
	    if atom(line) then
	        return EOF
	    else
	        if line[$] = 10 then
	            return line[1..$-1]
	        else
	            return line
	        end if
	    end if
	end function
    Method("GetS",routine_id("Input_GetS"))
--<li>[function] GetS() - reads a "line" from the Input Device
--<p> If the string ends in a '\n' then the linefeed will be removed.
--    If Ctrl-Z is pressed then -1 is returned.
--    If end-of-file reached then -1 is returned.
--</ul>
EndClass("Input")
--</dir>

Class("File")
--<li><h3><a name=File>"File"</a></h3>
--<dir>
--<h4>Parents: <a href=#Input>"Input"</a> & <a href=#Output>"Output"</a></h4>
    Extends({"Input","Output"})
--<h4>Properties:</h4>
--<ul>
    Property("NAME",STRING,"")
--<li>NAME - <code>STRING</code> string containing name of file
--</ul>
--<h4>Methods:</h4>
--<ul>
	procedure File_Close(Instance self)
	    close(MethodFunc(self,"Get",{"HANDLE"}))
	end procedure
	Method("Close",routine_id("File_Close"))
--<li>[procedure] Close() - closes the file
--</ul>
EndClass("File")
--<h4>Constructor:</h4>
global function File(string name,string mode)
    integer handle -- In Eu4 file_number handle
    Instance self
    self = New("File",{{"NAME",name}})
    handle = open(name,mode)
    if handle > 2 then
    MethodProc(self,"Set",{"HANDLE",handle})
    else
        printf(2,"**ERROR**\n*File '%s' cannot be opened*\n",{name})
    end if
    return self
end function
--[function] File(<code>string</code> name,<code>string</code> mode)
--<p>Parameters:
--<ul>
--<li><code>string</code> name - the name of the file you wish to open
--<li><code>string</code> mode - the mode of opening the file
--</ul>
-- Return value: a pointer to an Object of type "File"
--</dir>

--</ul>

--<h3>Defined Instances:</h3>
--<ul>
global constant
    StdOut = New("Screen",{{"HANDLE",1}}),
    StdErr = New("Screen",{{"HANDLE",2}}),
--<li><font color=purple>StdOut</font>
-- - the Output Object for the standard output device (Handle = 1)
--<li><font color=purple>StdErr</font>
-- - the Output Object for writing errors to the screen (Handle = 2)
    StdIn = New("Input",{{"HANDLE",0}})
--<li><font color=purple>StdIn</font>
-- - the Input Object for the standard input device (Handle = 0)
--</ul>

--<h3>Additional routines:</h3>
--<ul>
global procedure pause(string text)
    object wait
    MethodProc(StdOut,"Puts",{"\n\n" & text})
    wait = MethodFunc(StdIn,"GetC",{})
end procedure
--<li>[procedure] <font color=purple>pause</font>(<code>string</code> text)
-- - displays text on screen and awaits a key
-- depression
--</ul>

--</BODY>
--</HTML>
