--<HTML>
--<HEAD>
--<Title>File: Types.e</TITLE>
--</HEAD>
--<BODY>
--<h1>Types</h1>
-- This file contains some additional types for use in Euphoria programming.
--<p>It also contains some additional constants and routines which you might
-- find useful.
--<h2>Library specification</h2>
--<h3>Version: 3.0.0</h3>
--Date: September 2008
--<br>Author: C A Newbould (canewbould@yahoo.co.uk)

--<h2>Program Interface</h2>
  --<h3>Includes:</h3>
--<h3>Constants:</h3>
--<ul>
--<li>TRUE - set to 1
global constant TRUE = 1,
--<li>FALSE - set to 0
                FALSE = 0,
--<li>EMPTY_STRING - set to {}
                EMPTY_STRING = ""
--</ul>

--<h3>Routines:</h3>
--<ul>
--<li>[function] iif(boolean expression,object true,object false)
--<br>evaluates "expression" and returns "true" if TRUE or "false" otherwise
  -- ** defined below **

global function is_boolean(object bool)
    if not integer(bool) then
        return FALSE
    else
        if (bool = TRUE or bool = FALSE) then
            return TRUE
        else
            return FALSE
        end if
    end if
end function
--<li>[function] is_boolean(object bool)
--<br>returns TRUE if object is a boolean

global function is_byte(object byt)
    if not integer(byt) then
        return FALSE
    else
        if byt >= 0 and byt < 256 then
            return TRUE
        else
            return FALSE
        end if
    end if
end function
--<li>[function] is_byte(object byt)
--<br>returns TRUE if object is a byte (0 to 255)

global function is_character(object char)
    if not integer(char) then
        return FALSE
    else
        if char >= 0 and char < 128 then
            return TRUE
        else
            return FALSE
        end if
    end if
end function
--<li>[function] is_character(object char)
--<br>returns TRUE if object is a character (0 to 127)

--<li>[function] is_array(sequence arr)
--<br>returns TRUE if "arr" is an array
  -- ** defined below **

--<li>[function] is_string(sequence str)
--<br>returns TRUE if "str" is a string
  -- ** defined below **

--<li>[function] is_array_of_integers(sequence arr)
--<br>returns TRUE if "arr" is an array of integers
  -- ** defined below **

--<li>[function] is_array_of_strings(sequence arr)
--<br>returns TRUE if "arr" is an array of strings
  -- ** defined below **
--</ul>

--<h3>Types:</h3>
--<ul>

global type boolean(object b)
	return is_boolean(b)
end type
--<li>boolean - TRUE or FALSE

global type byte(object byt)
    return is_byte(byt)
end type
--<li>byte - 0 to 255

global type character(object char)
    return is_character(char)
end type
--<li>character - 0 to 127

global type alphanum_char(object x)
   if integer(x) then
      if ('A' <= x and x <= 'Z')
      or ('a' <= x and x <= 'z')
      or ('0' <= x and x <= '9') then
         return 1
      end if
   end if
   return 0
end type
--<li>alphanum_char - character in range: 48-57, 65-90, 97-122

  -- Routines defined once "boolean" defined
global function iif(boolean expression,object true,object false)
    if expression then
        return true
    else
        return false
    end if
end function

function get_type(object obj)
    sequence mask
    mask = repeat(FALSE,3)
    if atom(obj) then
        mask[1] = TRUE
    end if
    if integer(obj) then
        mask[2] = TRUE
    end if
    if sequence(obj) then
        mask[3] = TRUE
    end if
    return mask
end function

global function is_array(object a)
	object first
	integer len
	boolean	ret
	if not sequence(a) then
		return FALSE
	else
		ret = TRUE
		len = length(a)
		if len > 0 then
			first = get_type(a[1])
			for i = 2 to len do
				if not equal(get_type(a[i]),first) then
					ret = FALSE
				end if
			end for
		end if
	end if
	return ret
end function

global type array(sequence arr)
    return is_array(arr)
end type
--<li>array - a one-dimensional array of compatible type (all <code>atom</code>,
-- all <code>integer</code> or all <code>sequence</code>)
-- (There is a slight problem here - Euphoria's <code>integer</code> function
-- doesn't distinguish between x and x.0!)
  -- Can I address this?

global function is_string(object str)
    boolean result
    if not is_array(str) then
        result = FALSE
    else
        result = TRUE
        for i = 1 to length(str) do
            if not is_character(str[i]) then
                result = FALSE
            end if   
        end for
    end if
    return result
end function

global type string(sequence str)
    return is_string(str)
end type
--<li>string - a one-dimensional array of characters
-- including the empty string {}

global function is_extendedstring(object str)
    boolean result
    if not is_array(str) then
        result = FALSE
    else
        result = TRUE
        for i = 1 to length(str) do
            if not is_byte(str[i]) then
                result = FALSE
            end if
        end for
    end if
    return result
end function

global type extended_string(sequence str)
    return is_extendedstring(str)
end type
--<li>extended_string - a one-dimensional array of bytes
-- including the empty string {}

global function is_array_of_integers(object arr)
    boolean result
    if not is_array(arr) then
        result = FALSE
    else
        result = TRUE
        for i = 1 to length(arr) do
            if not integer(arr[i]) then
                result = FALSE
            end if
        end for
    end if
    return result
end function

global type array_of_integers(sequence arr)
   return is_array_of_integers(arr)
end type
--<li>array_of_integers - a one-dimensional array of integers

global function is_array_of_strings(object arr)
    boolean result
    if not is_array(arr) then
        result = FALSE
    else
        result = TRUE
        for i = 1 to length(arr) do
            if not is_string(arr[i]) then
                result = FALSE
            end if
        end for
    end if
    return result
end function

global type array_of_strings(sequence arr)
   return is_array_of_strings(arr)
end type
--<li>array_of_strings - a one-dimensional array of strings

--</ul>

--</BODY>
--</HTML>



