--------------------------------------------------------------------------------
-- Name:            Library gregEaster - Gregorian Easter

-- Version:		    .e

-- Author:          C. A. Newbould

-- Created:         August 2006

-- Notes:
--      Contains a single function
--------------------------------------------------------------------------------

-- Routines
	
--------------------------------------------------------------------------------
-- Method: gregEaster
-- Usage:  gregEaster(integer year)
-- Parameter: any year after 1582
-- Purpose: uses Gauss's algorithm to locate Easter Day for the given year,
--          based on the Gregorian calendar.
-- Return value: a sequence containing:
-- 			     [1] day of month in "%d" format (i.e. 1 - 31);
--               [2] month in "%d" format (i.e. 3 or 4).
-- Scope: global
-- Note:
-- Currently defined up to 2299  - put these bounds into the calling program.
--------------------------------------------------------------------------------
global function gregEaster(integer year)
	
	integer a, b, c, d, e, f, m, n, dd, mm

	a = remainder(year,19)
	b = remainder(year,4)
	c = remainder(year,7)
	
	m = 22
	n = 2
	f = floor(year/100)
	if  f = 17 or f = 18 then 
		m += 1
		n += (f - 16)
	elsif f = 19 or f = 20 then
		m += 2
		n += 3
	elsif f = 21 then
		m += 2
		n += 4
	elsif f = 22 then
		m += 3
		n = 0	-- "plus 5" yielding 7 modulo 7
	end if
		
	d = remainder(19*a + m,30)
	e = remainder(2*b + 4*c + 6*d + n,7)
	
	if (d + e) < 10 then
		dd = 22 + d + e
		mm = 3
	else
		dd = d + e - 9
		mm = 4
		-- deal with the two exceptions
		if dd = 26 or (dd = 25 and d = 28 and e = 6 and a > 10) then
			dd -= 7		
		end if
	end if
	
	return {dd,mm}
	
end function

--------------------------------------------------------------------------------

-- Example Program: TestEaster.ex

--------------------------------------------------------------------------------

-- end of Library gregEaster.e

--------------------------------------------------------------------------------