-- CommonTypes.e ------------------------------------------------------------

-- Some types of common interest in programming. Incorporates types from old
-- (deprecated) file EasyType.e

-- MATHEMATICAL -------------------------------------------------------------
-- Defines types:
--	bool

-- --------------------------------------------------------------------------
-- FredMangan6@gmail.com

global type bool(object i)
	if integer(i) then
		return i=1 or i=0
	end if
	return 0
end type -- bool ------------------------------------------------------------

global type even (object a)
-- is a an even number?
	return atom(a) and floor(a/2)*2 = a
end type

-- GRAPHICAL and COORD-GEOM *************************************************
-- (including types used for my drawing programs)
-- Defines types:
--	num_pair	point	polyline	r_diag		polypic_element		polypic_deprecated

-- Definition of type coord() caused a type error in
-- SScale.e at line 137, even though coord there is a local var.
-- So define type num_pair(), also point(), instead:
global type num_pair(object s)

	if sequence(s) and length(s) = 2 then
		if atom(s[1]) and atom(s[2]) then
			return 1 -- ok
		end if
	end if
	
	return 0
end type -- num_pair --------------------------------------------------------

global type point(sequence s)
-- Define euphoria expression of a point.
	return num_pair(s)
end type -- point

global type polyline(object s)
-- A polyline is a sequence of at least 1 num_pairs
    if sequence(s) and length(s) then
        for i=1 to length(s) do
            if not num_pair(s[i]) then
                return 0    -- fails
            end if
        end for
        return 1            -- ok
    end if
    return 0                -- fails
end type -- polyline --------------------------------------------------------

global type r_diag(object s)
-- Many euphoria routines require a rising diagonal.
-- This defines it.
    if sequence(s) and length(s) = 2 and point(s[1]) and point(s[2]) then
    	if (s[1][1] < s[2][1]) and (s[1][2] < s[2][2]) then
        	return 1
    	end if
    end if
    
    return 0
end type -- r_diag ----------------------------------------------------------

global type polypic_element(object s)
    
    if sequence(s) and length(s) = 4 then
        if atom(s[1]) and atom(s[2]) and atom(s[3]) and sequence(s[4]) then
            return 1
        end if
    end if
    
    return 0
end type -- polypic_element -------------------------------------------------

-- ALPHA-NUMERICAL**********************************************************
-- Defines types:
--	ascii		nul_string		ascii_string		alpha		num	alphanum
--	fieldname

global type ascii(object a)
-- Define an ascii character
    return sequence(a) and a >= 0 and a <= 255
end type -- ascii -----------------------------------------------------------

global type nul_string(object s)
	return sequence(s) and equal(s,"")
end type -- nul_string

global type ascii_string(object s)
-- Defines an ascii string - includes a null string
    if sequence(s) then
    	if nul_string(s) then
    		return 1
    	else
    		for i = 1 to length(s) do
        		if atom(s[i]) then
        			if not ascii(s[i]) then
            			return 0
        			end if
        		else
        			return 0	-- not an atom
        		end if
    		end for
    		return 1
    	end if
    end if
    return 0 -- not a sequence
end type -- ascii_string ----------------------------------------------------

global type alpha(object a)
-- Is a character alphabetical?
    if integer(a) and (a >= 'A' and a <= 'Z') or (a >= 'a' and a <= 'z') then
    	return 1
    end if
    return 0
end type -- alpha -----------------------------------------------------------

global type num(object a)
-- Is a character numeric?
    if integer(a) and a >= '0' and a <= '9' then
    	return 1
    end if
    return 0
end type  -- num ------------------------------------------------------------

global type alphanum(object a)
-- Is a character alphanumerical?
    return integer(a) and (alpha(a) or num(a))
end type -- alphanum --------------------------------------------------------

global type fieldname(sequence s)
-- Used to define a data_base fieldname
-- To be valid, s should be a string containing only alphanumeric characters
-- or underscores.
	
    if sequence(s) then
    	if length(s) = 0 then
    		return 0			-- it is an empty string
    	else
    		for i = 1 to length(s) do
    			if integer(s[i]) then
        			if not (alphanum(s[i]) or s[i] = '_') then
        				return 0
        			end if
    			else
    				return 0
    			end if
    		end for
    		return 1			-- it is valid, if we get to here
    	end if
    end if
    
    return 0
end type -- fieldname -------------------------------------------------------

